#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
import os
import platform
import sys
from ctypes.util import find_library
from importlib.metadata import PackageMetadata
from pathlib import Path
from typing import Optional

import bitcoin_usb
import bitcointx

from .html import link
from .i18n import translate

logger = logging.getLogger(__name__)


from PyQt6.QtWidgets import QApplication, QMessageBox


# Function to show the warning dialog before starting the QApplication
def show_message_before_quit(msg: str) -> None:
    # Initialize QApplication first
    app = QApplication(sys.argv)
    # Without an application instance, some features might not work as expected
    QMessageBox.warning(None, "Warning", msg, QMessageBox.StandardButton.Ok)  # type: ignore[arg-type]
    sys.exit(app.exec())


def get_libsecp256k1_electrumsv_path() -> str:
    from electrumsv_secp256k1 import _libsecp256k1

    # Get the platform-specific path to the binary library
    if platform.system() == "Windows":
        # On Windows, construct the path to the DLL
        here = os.path.dirname(os.path.abspath(_libsecp256k1.__file__))
        lib_path = os.path.join(here, "libsecp256k1.dll")
    else:
        # On Linux and macOS, directly use the __file__ attribute
        lib_path = _libsecp256k1.__file__
    return lib_path


def get_libsecp256k1_os_path() -> str | None:
    "This cannot be used directly, because it doesnt return an absolute path"
    lib_name = "secp256k1"
    return find_library(lib_name)


def get_packaged_libsecp256k1_path() -> str | None:
    # for apppimage it is
    # __file__ = squashfs-root/usr/lib/python3.10/site-packages/bitcoin_safe/dynamic_lib_load.py
    # and the lib is in
    # squashfs-root/usr/lib/libsecp256k1.so.0.0.0
    packaged_lib_path = Path(__file__).parent.parent.parent.parent
    for name in ["libsecp256k1.so.0.0.0", "libsecp256k1.so.0"]:
        lib_path = packaged_lib_path / name
        if lib_path.exists():
            return str(lib_path)
    return None


def setup_libsecp256k1() -> None:
    """The operating system might, or might not provide libsecp256k1 needed for bitcointx

    Therefore we require https://pypi.org/project/electrumsv-secp256k1/ in the build process as additional_requires
    and point the bicointx library here to this binary.

    This isn't ideal, but:
        # electrumsv-secp256k1 offers libsecp256k1 prebuild for different platforms
        # which is needed for bitcointx.
        # bitcointx and with it the prebuild libsecp256k1 is not used for anything security critical
        # key derivation with bitcointx is restricted to testnet/regtest/signet
        # and the PSBTTools using bitcointx is safe because it handles no key material
    """

    lib_path = None

    # 1 choice is the packaged version
    packaged_libsecp256k1_path = get_packaged_libsecp256k1_path()
    if packaged_libsecp256k1_path:
        logger.info(f"libsecp256k1 found in package.: {packaged_libsecp256k1_path}")
        lib_path = packaged_libsecp256k1_path

    # Fallback choice is the electrumsv version
    if not lib_path:
        binary_lib_path_from_electrumsv = get_libsecp256k1_electrumsv_path()
        if binary_lib_path_from_electrumsv:
            logger.info(f"libsecp256k1 found via fallbackmethod: {binary_lib_path_from_electrumsv}")
            lib_path = binary_lib_path_from_electrumsv

    if lib_path:
        logger.info(f"Setting libsecp256k1: {lib_path}")
        bitcoin_usb.set_custom_secp256k1_path(lib_path)
        bitcointx.set_custom_secp256k1_path(lib_path)
    else:
        logger.info(f"libsecp256k1 could not be found at all. This app will not start")


def ensure_pyzbar_works() -> None:
    "Ensure Visual C++ Redistributable Packages for Visual Studio 2013"
    # Get the platform-specific path to the binary library
    logger.info(f"Platform: {platform.system()}")
    if platform.system() == "Windows":
        logger.info("Trying to import pyzbar to see if Visual C++ Redistributable is installed. ")
        try:
            from pyzbar import pyzbar

            pyzbar.__name__
            logger.info(f"pyzbar successfully loaded ")
        except FileNotFoundError:
            logger.info(f"pyzbar not loaded ")
            show_message_before_quit(
                translate("lib_load", """You are missing the {link}\nPlease install it.""").format(
                    link=link(
                        "https://www.microsoft.com/en-US/download/details.aspx?id=40784",
                        "Visual C++ Redistributable Packages for Visual Studio 2013",
                    )
                ),
            )
            sys.exit()
    else:
        # On Linux this shoudn't be a problem, because I specidfied
        # system_runtime_requires = [ "libzbar0",  ....#
        pass


def get_briefcase_meta_data() -> Optional[PackageMetadata]:
    import sys
    from importlib import metadata as importlib_metadata

    # Find the name of the module that was used to start the app
    app_module = sys.modules["__main__"].__package__
    if not app_module:
        return None
    # Retrieve the app's metadata
    metadata = importlib_metadata.metadata(app_module)

    return metadata



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
import os
import platform
import sys

from bitcoin_safe import __version__

from .simple_mailer import compose_email


def remove_absolute_paths(line: str) -> str:
    """
    Replaces absolute paths in a traceback line with relative ones,
    based on the current script's execution path.
    """
    current_path = os.getcwd() + os.sep
    return line.replace(current_path, "")


def mail_error_repot(error_report: str) -> None:
    email = "andreasgriffin@proton.me"
    subject = f"Error report - Bitcoin Safe Version: {__version__}"
    body = f"""Error:
            {error_report}
            """.replace(
        "    ", ""
    )

    # Write additional system info if needed
    body += "\n\nSystem Info:\n"
    body += f"OS: {platform.platform()}\n"
    body += f"Python Version: {sys.version}\n"
    body += f"Bitcoin Safe Version: {__version__}\n\n"
    return compose_email(email, subject, body)


class RelativePathFormatter(logging.Formatter):
    def formatException(self, exc_info) -> str:
        return remove_absolute_paths(super().formatException(exc_info))

    def format(self, record) -> str:
        if record.exc_info:
            record.exc_text = self.formatException(record.exc_info)
        return super().format(record)


class MailHandler(logging.Handler):
    def __init__(self, level=logging.NOTSET, must_include_exc_info=True) -> None:
        super().__init__(level)
        self.must_include_exc_info = must_include_exc_info

    def emit(self, record) -> None:
        """'args', 'asctime', 'created', 'exc_info', 'exc_text', 'filename', 'funcName', 'getMessage', 'levelname', 'levelno', 'lineno', 'message', 'module', 'msecs', 'msg', 'name', 'pathname', 'process', 'processName', 'relativeCreated', 'stack_info', 'thread', 'threadName"""

        if (self.must_include_exc_info and record.exc_info) or not self.must_include_exc_info:
            exc_type, exc_value, exc_traceback = record.exc_info  # type: ignore

            message = str(self.format(record))
            mail_error_repot(message)



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from typing import List

from PyQt6 import QtGui
from PyQt6.QtCore import QStringListModel, Qt, pyqtSignal
from PyQt6.QtGui import QKeyEvent
from PyQt6.QtWidgets import (
    QApplication,
    QCompleter,
    QHBoxLayout,
    QLineEdit,
    QSizePolicy,
    QStyle,
    QVBoxLayout,
    QWidget,
)

from bitcoin_safe.gui.qt.category_list import CategoryEditor

logger = logging.getLogger(__name__)


class LabelLineEdit(QLineEdit):
    signal_enterPressed = pyqtSignal()  # Signal for Enter key
    signal_textEditedAndFocusLost = pyqtSignal()  # Signal for text edited and focus lost

    def __init__(self, parent=None):
        super().__init__(parent)
        self.originalText = ""
        self.textChangedSinceFocus = False
        self.installEventFilter(self)  # Install an event filter

        self._model = QStringListModel()
        self._completer = QCompleter(self._model, self)
        self._completer.setCompletionMode(QCompleter.CompletionMode.UnfilteredPopupCompletion)
        self._completer.setFilterMode(Qt.MatchFlag.MatchContains)
        self._completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self.setCompleter(self._completer)

        # signals
        self.textChanged.connect(self.onTextChanged)  # Connect the textChanged signal

    def set_completer_list(self, strings: List[str]):
        self._model.setStringList(strings)
        self._completer.setModel(self._model)

    def onTextChanged(self):
        self.textChangedSinceFocus = True  # Set flag when text changes

    def eventFilter(self, obj, event):
        if obj == self:
            if event.type() == QKeyEvent.Type.FocusIn:
                self.originalText = self.text()  # Store text when focused
                self.textChangedSinceFocus = False  # Reset change flag
            elif event.type() == QKeyEvent.Type.FocusOut:
                if self.textChangedSinceFocus:
                    self.signal_textEditedAndFocusLost.emit()  # Emit signal if text was edited
                self.textChangedSinceFocus = False  # Reset change flag
        return super().eventFilter(obj, event)

    def keyPressEvent(self, event: QKeyEvent | None):
        if not event:
            super().keyPressEvent(event)
            return

        if event.key() == Qt.Key.Key_Return or event.key() == Qt.Key.Key_Enter:
            self.signal_enterPressed.emit()  # Emit Enter pressed signal
        elif event.key() == Qt.Key.Key_Escape:
            self.setText(self.originalText)  # Reset text on ESC
        elif self._model.stringList() and event.key() in (Qt.Key.Key_Up, Qt.Key.Key_Down):
            popup = self._completer.popup()
            if popup and not popup.isVisible():
                self._completer.complete()
        else:
            super().keyPressEvent(event)


class LabelAndCategoryEdit(QWidget):
    def __init__(
        self,
        parent=None,
        dismiss_label_on_focus_loss=False,
    ) -> None:
        super().__init__(parent=parent)
        self.label_edit = LabelLineEdit(parent=self)
        self.category_edit = QLineEdit(parent=self)
        self.category_edit.setReadOnly(True)
        self.category_edit.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.category_edit.setAlignment(Qt.AlignmentFlag.AlignCenter)
        # self.category_edit.setFixedWidth(100)

        self.main_layout = QHBoxLayout(
            self
        )  # Horizontal layout to place the input field and buttons side by side

        # Add the input field and buttons layout to the main layout
        self.main_layout.addWidget(self.category_edit)
        self.main_layout.addWidget(self.label_edit)

        # Ensure there's no spacing that could affect the alignment
        self.main_layout.setSpacing(0)
        self.main_layout.setContentsMargins(0, 0, 0, 0)

        if dismiss_label_on_focus_loss:
            self.label_edit.signal_textEditedAndFocusLost.connect(
                lambda: self.label_edit.setText(self.label_edit.originalText)
            )

    def _format_category_edit(self) -> None:
        palette = QtGui.QPalette()
        background_color = None

        if self.category_edit.text():
            background_color = CategoryEditor.color(self.category_edit.text())
            palette.setColor(QtGui.QPalette.ColorRole.Base, background_color)
        else:
            palette = (self.category_edit.style() or QStyle()).standardPalette()

        self.category_edit.setPalette(palette)
        self.category_edit.update()

    def set(self, label: str, category: str):

        self.set_label(label)
        self.set_category(category)

    def set_category(self, category: str):

        self.category_edit.setText(category)
        self._format_category_edit()

    def set_label(
        self,
        label: str,
    ):
        self.label_edit.setText(label)
        self.label_edit.originalText = label

    def set_placeholder(
        self,
        text: str,
    ):
        self.label_edit.setPlaceholderText(text)

    def set_category_visible(self, value: bool):

        self.category_edit.setVisible(value)

    def category(self) -> str:
        return self.category_edit.text()

    def label(self) -> str:
        return self.label_edit.text().strip()

    def set_label_readonly(self, value: bool):
        self.label_edit.setReadOnly(value)


# Example usage
if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    widget = QWidget()
    widget.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
    widget_layout = QVBoxLayout(widget)
    widget_layout.setContentsMargins(0, 0, 0, 0)
    widget_layout.setSpacing(0)

    edit = LabelAndCategoryEdit()
    edit.set("some label", "KYC")
    widget_layout.addWidget(edit)

    widget.show()
    sys.exit(app.exec())



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import enum
import logging
from math import ceil
from typing import Any, Dict, List, Optional, Tuple

from bitcoin_safe.gui.qt.util import custom_exception_handler
from bitcoin_safe.network_config import NetworkConfig
from bitcoin_safe.signals import SignalsMin
from bitcoin_safe.threading_manager import TaskThread, ThreadingManager

from .config import MIN_RELAY_FEE

logger = logging.getLogger(__name__)

import datetime

import numpy as np
import requests
from PyQt6.QtCore import QObject, pyqtSignal

feeLevels = [
    1,
    2,
    3,
    4,
    5,
    6,
    8,
    10,
    12,
    15,
    20,
    30,
    40,
    50,
    60,
    70,
    80,
    90,
    100,
    125,
    150,
    175,
    200,
    250,
    300,
    350,
    400,
    500,
    600,
    700,
    800,
    900,
    1000,
    1200,
    1400,
    1600,
    1800,
    2000,
]
chartColors = [
    "#D81B60",
    "#8E24AA",
    "#5E35B1",
    "#3949AB",
    "#1E88E5",
    "#039BE5",
    "#00ACC1",
    "#00897B",
    "#43A047",
    "#7CB342",
    "#C0CA33",
    "#FDD835",
    "#FFB300",
    "#FB8C00",
    "#F4511E",
    "#6D4C41",
    "#757575",
    "#546E7A",
    "#b71c1c",
    "#880E4F",
    "#4A148C",
    "#311B92",
    "#1A237E",
    "#0D47A1",
    "#01579B",
    "#006064",
    "#004D40",
    "#1B5E20",
    "#33691E",
    "#827717",
    "#F57F17",
    "#FF6F00",
    "#E65100",
    "#BF360C",
    "#3E2723",
    "#212121",
    "#263238",
    "#801313",
]


mempoolFeeColors = [
    "#557d00",
    "#5d7d01",
    "#637d02",
    "#6d7d04",
    "#757d05",
    "#7d7d06",
    "#867d08",
    "#8c7d09",
    "#957d0b",
    "#9b7d0c",
    "#a67d0e",
    "#aa7d0f",
    "#b27d10",
    "#bb7d11",
    "#bf7d12",
    "#bf7815",
    "#bf7319",
    "#be6c1e",
    "#be6820",
    "#bd6125",
    "#bd5c28",
    "#bc552d",
    "#bc4f30",
    "#bc4a34",
    "#bb4339",
    "#bb3d3c",
    "#bb373f",
    "#ba3243",
    "#b92b48",
    "#b9254b",
    "#b8214d",
    "#b71d4f",
    "#b61951",
    "#b41453",
    "#b30e55",
    "#b10857",
    "#b00259",
    "#ae005b",
]


def fee_to_color(fee, colors=chartColors) -> str:
    if fee == 0:
        # for 0 just use the same color as 1
        fee = 1
    indizes = np.where(np.array(feeLevels) <= fee)[0]
    if len(indizes) == 0:
        return "#000000"
    return colors[indizes[-1]]


def fetch_from_url(url: str, is_json=True) -> Optional[Any]:
    logger.debug(f"fetch_json_from_url requests.get({url}, timeout=10)")

    try:
        response = requests.get(url, timeout=10)
        # Check if the request was successful (status code 200)
        if response.status_code == 200:
            # Parse the JSON response
            data = response.json() if is_json else response.content
            return data
        else:
            # If the request was unsuccessful, print the status code
            logger.error(f"Request failed with status code: {response.status_code}")
            return None
    except:
        logger.error(f"fetch_json_from_url {url} failed")
        return None


def threaded_fetch(url: str, on_success, is_json=True) -> TaskThread:
    def do() -> Any:
        return fetch_from_url(url, is_json=is_json)

    def on_error(packed_error_info) -> None:
        custom_exception_handler(*packed_error_info)

    def on_done(data) -> None:
        pass

    return TaskThread().add_and_start(do, on_success, on_done, on_error)


class TxPrio(enum.Enum):
    low = enum.auto()
    medium = enum.auto()
    high = enum.auto()


class MempoolData(ThreadingManager, QObject):
    signal_data_updated = pyqtSignal()

    def __init__(
        self,
        network_config: NetworkConfig,
        signals_min: SignalsMin,
        threading_parent: ThreadingManager,
    ) -> None:
        super().__init__(threading_parent=threading_parent)
        self.signals_min = signals_min

        self.network_config = network_config
        self.mempool_blocks = self._empty_mempool_blocks()
        self.recommended: Dict[str, int] = {
            "fastestFee": MIN_RELAY_FEE,
            "halfHourFee": MIN_RELAY_FEE,
            "hourFee": MIN_RELAY_FEE,
            "economyFee": MIN_RELAY_FEE,
            "minimumFee": MIN_RELAY_FEE,
        }
        self.time_of_data = datetime.datetime.fromtimestamp(0)
        self.mempool_dict: Dict[str, Any] = {
            "count": 0,
            "vsize": 0,
            "total_fee": 0,
            "fee_histogram": [],
        }
        logger.debug(f"initialized {self}")

    def _empty_mempool_blocks(self) -> List[Dict[str, Any]]:
        return [
            {
                "blockSize": 1,
                "blockVSize": 1,
                "nTx": 0,
                "totalFees": MIN_RELAY_FEE,
                "medianFee": MIN_RELAY_FEE,
                "feeRange": [MIN_RELAY_FEE, MIN_RELAY_FEE],
            }
        ]

    def fee_rates_min_max(self, block_index: int) -> Tuple[int, float]:
        block_index = min(block_index, len(self.mempool_blocks) - 1)
        fee_rates = self.mempool_blocks[block_index]["feeRange"]
        return min(fee_rates), max(fee_rates)

    def median_block_fee_rate(self, block_index: int) -> float:
        block_index = min(block_index, len(self.mempool_blocks) - 1)
        return self.mempool_blocks[block_index]["medianFee"]

    def num_mempool_blocks(self) -> int:
        vBytes_per_block = 1e6
        return ceil(self.mempool_dict["vsize"] / vBytes_per_block)

    def get_prio_fee_rates(self) -> Dict[TxPrio, float]:
        return {
            prio: self.recommended[key]
            for prio, key in zip(
                [TxPrio.high, TxPrio.medium, TxPrio.low],
                ["fastestFee", "halfHourFee", "hourFee"],
            )
        }

    def get_min_relay_fee_rate(self) -> float:
        return self.recommended["minimumFee"]

    def max_reasonable_fee_rate(self) -> float:
        "Average fee of the 0 projected block"
        average_fee_rate = sum(self.fee_rates_min_max(0)) / 2

        # allow for up to 20% more then the average_fee_rate
        slack = 0.2

        return average_fee_rate * (1 + slack)

    def set_data_from_mempoolspace(self, force=False) -> None:
        if not force and datetime.datetime.now() - self.time_of_data < datetime.timedelta(minutes=9):
            logger.debug(
                f"Do not fetch data from {self.network_config.mempool_url} because data is only {datetime.datetime.now()- self.time_of_data  } old."
            )
            return None

        self.time_of_data = datetime.datetime.now()

        def on_mempool_blocks(mempool_blocks) -> None:
            if mempool_blocks:
                self.mempool_blocks = mempool_blocks
                logger.info(f"Updated mempool_blocks {mempool_blocks}")

        self.append_thread(
            threaded_fetch(
                f"{self.network_config.mempool_url}api/v1/fees/mempool-blocks",
                on_mempool_blocks,
            )
        )
        logger.debug(f"started on_mempool_blocks")

        def on_recommended(recommended) -> None:
            if recommended:
                self.recommended = recommended
                logger.info(f"Updated recommended {recommended}")

        self.append_thread(
            threaded_fetch(
                f"{self.network_config.mempool_url}api/v1/fees/recommended",
                on_recommended,
            )
        )
        logger.debug(f"started on_recommended")

        def on_mempool_dict(mempool_dict) -> None:
            if mempool_dict:
                self.mempool_dict = mempool_dict
                logger.info(f"Updated mempool_dict {mempool_dict}")
            self.signal_data_updated.emit()

        self.append_thread(
            threaded_fetch(
                f"{self.network_config.mempool_url}api/mempool",
                on_mempool_dict,
            )
        )
        logger.debug(f"started on_mempool_dict")

    def fetch_block_tip_height(self) -> int:
        response = fetch_from_url(f"{self.network_config.mempool_url}api/blocks/tip/height")
        return response if response else 0

    def fee_rate_to_projected_block_index(self, fee_rate: float) -> int:
        available_blocks = len(self.mempool_blocks)
        for i in range(available_blocks):
            v_min, v_max = self.fee_rates_min_max(i)
            if fee_rate >= v_min:
                return i
        return available_blocks



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import datetime
import logging
import random
import sys

import numpy as np
from PyQt6.QtCharts import QChart, QChartView, QDateTimeAxis, QLineSeries, QValueAxis
from PyQt6.QtCore import QDateTime, QMargins, Qt, QTimer
from PyQt6.QtWidgets import QApplication, QFrame, QMainWindow, QVBoxLayout, QWidget

from bitcoin_safe.util import unit_str

from ...signals import UpdateFilter, WalletSignals
from ...wallet import Wallet

logger = logging.getLogger(__name__)


class BalanceChart(QWidget):
    def __init__(self, y_axis_text="Balance") -> None:
        super().__init__()
        self.y_axis_text = y_axis_text

        # Layout
        layout = QVBoxLayout()

        # Create chart
        self.chart = QChart()
        self.chart.setBackgroundBrush(Qt.GlobalColor.white)
        if legend := self.chart.legend():
            legend.hide()

        # Reduce the overall chart margins
        layout.setContentsMargins(QMargins(0, 0, 0, 0))  # Smaller margins (left, top, right, bottom)
        self.chart.setMargins(QMargins(0, 0, 0, 0))  # Smaller margins (left, top, right, bottom)

        # Create DateTime axis for X
        self.datetime_axis = QDateTimeAxis()

        # Create Value axis for Y
        self.value_axis = QValueAxis()

        self.chart.addAxis(self.datetime_axis, Qt.AlignmentFlag.AlignBottom)
        self.chart.addAxis(self.value_axis, Qt.AlignmentFlag.AlignLeft)
        self.chart.setBackgroundRoundness(0)

        # Add chart to chart view
        self.chart_view = QChartView(self.chart)
        layout.addWidget(self.chart_view)

        self.chart_view.setFrameStyle(QFrame.Shape.NoFrame)
        self.chart_view.setBackgroundBrush(self.chart.backgroundBrush())

        # Set layout
        self.setLayout(layout)

    def set_value_axis_label_format(self, max_value) -> None:
        if max_value != 0:
            # Determine the number of digits before the decimal
            import math

            magnitude = int(math.log10(abs(max_value)))
            decimals = -min(magnitude, 0) + 1

        else:
            # Default to three decimal places if max_value is 0
            decimals = 3

        # Set the label format on the axis
        format_string = f"%.{decimals}f"
        self.value_axis.setLabelFormat(format_string)

    def update_chart(self, balance_data, project_until_now=True) -> None:
        if len(balance_data) == 0:
            return
        balance_data = np.array(balance_data).copy()

        self.datetime_axis.setTitleText(self.tr("Date"))
        self.datetime_axis.setTickCount(6)
        self.value_axis.setTitleText(self.y_axis_text)
        self.set_value_axis_label_format(np.max(balance_data[:, 1]))

        # Clear previous series
        self.chart.removeAllSeries()

        # Variables to store the min/max values for the axes
        min_balance = 0
        max_balance = 0
        min_timestamp = float("inf")
        max_timestamp = float("-inf")
        for timestamp, balance in balance_data:
            max_balance = max(max_balance, balance)
            min_balance = min(min_balance, balance)
            max_timestamp = max(max_timestamp, timestamp)
            min_timestamp = min(min_timestamp, timestamp)

        #  add the 0 balance as first data point
        balance_data = np.vstack(
            [
                (min_timestamp, min_balance),
                balance_data,
            ]
        )

        # add the current time as last point, if the data is in the past
        if project_until_now and datetime.datetime.now().timestamp() > max(balance_data[:, 0]):
            balance_data = np.vstack(
                [
                    balance_data,
                    (datetime.datetime.now().timestamp(), balance_data[-1][1]),
                ]
            )

        if np.max(balance_data[:, 0]) - np.min(balance_data[:, 0]) < 24 * 60 * 60:
            self.datetime_axis.setFormat("HH:mm")
        else:
            self.datetime_axis.setFormat("d MMM yy")

        # Create Line series
        series = QLineSeries()

        for i, (timestamp, balance) in enumerate(balance_data[:-1]):
            next_timestamp, _ = balance_data[i + 1]
            series.append(
                QDateTime.fromSecsSinceEpoch(int(timestamp)).toMSecsSinceEpoch(),
                balance,
            )
            series.append(
                QDateTime.fromSecsSinceEpoch(int(next_timestamp)).toMSecsSinceEpoch(),
                balance,
            )

        # Add the last data point
        timestamp, balance = balance_data[-1]
        series.append(QDateTime.fromSecsSinceEpoch(int(timestamp)).toMSecsSinceEpoch(), balance)

        self.datetime_axis.setRange(
            QDateTime.fromSecsSinceEpoch(int(min_timestamp)),
            QDateTime.fromSecsSinceEpoch(int(max_timestamp)),
        )
        self.value_axis.setRange(
            0,
            max_balance,
        )

        buffer_time = 60 * 60
        self.datetime_axis.setMin(QDateTime.fromSecsSinceEpoch(int(min_timestamp - buffer_time)))
        self.datetime_axis.setMax(QDateTime.fromSecsSinceEpoch(int(max_timestamp + buffer_time)))
        self.value_axis.setMin(min_balance)
        buffer_factor = 0.01
        self.value_axis.setMax(max_balance * (1 + buffer_factor))

        # Adding series to the chart
        self.chart.addSeries(series)

        # Attach axes
        series.attachAxis(self.datetime_axis)
        series.attachAxis(self.value_axis)

        print(f"Actual DateTime Axis Range: {self.datetime_axis.min()} - {self.datetime_axis.max()}")

        # scatter_series = QScatterSeries()
        # for (timestamp, balance) in balance_data:
        #     scatter_series.append(
        #         QDateTime.fromSecsSinceEpoch(int(timestamp)).toMSecsSinceEpoch(),
        #         balance,
        #     )

        # scatter_series = QScatterSeries()
        # for (timestamp, balance) in balance_data:
        #     scatter_series.append(
        #         QDateTime.fromSecsSinceEpoch(int(timestamp)).toMSecsSinceEpoch(),
        #         balance,
        #     )

        # # Set marker shape and size
        # scatter_series.setMarkerShape(QScatterSeries.MarkerShape.MarkerShapeCircle)
        # scatter_series.setMarkerSize(5)

        # # Set marker color
        # border_color = QColor("#209fdf")  # Blue
        # brush_color = QColor("#209fdf")  # Semi-transparent blue
        # scatter_series.setPen(border_color)
        # scatter_series.setBrush(brush_color)

        # self.chart.addSeries(scatter_series)
        # scatter_series.attachAxis(self.datetime_axis)
        # scatter_series.attachAxis(self.value_axis)


class WalletBalanceChart(BalanceChart):
    def __init__(self, wallet: Wallet, wallet_signals: WalletSignals) -> None:
        super().__init__(y_axis_text="")
        self.value_axis.setLabelFormat("%.2f")
        self.wallet = wallet
        self.wallet_signals = wallet_signals

        self.updateUi()

        # signals
        self.wallet_signals.updated.connect(self.update_balances)
        self.wallet_signals.language_switch.connect(self.updateUi)

    def updateUi(self) -> None:
        self.y_axis_text = self.tr("Balance ({unit})").format(unit=unit_str(self.wallet.network))

        self.datetime_axis.setTitleText(self.tr("Date"))
        self.value_axis.setTitleText(self.y_axis_text)
        self.chart.update()

    def update_balances(self, update_filter: UpdateFilter) -> None:
        should_update = False
        if should_update or update_filter.refresh_all:
            should_update = True
        if should_update or update_filter.outpoints:
            should_update = True

        if not should_update:
            return

        logger.debug(f"{self.__class__.__name__} update_with_filter {update_filter}")

        # Calculate balance
        balance = 0
        balance_data = []
        for transaction_details in self.wallet.sorted_delta_list_transactions():
            balance += transaction_details.received - transaction_details.sent
            time = (
                transaction_details.confirmation_time.timestamp
                if transaction_details.confirmation_time
                else datetime.datetime.now().timestamp()
            )
            balance_data.append((time, balance / 1e8))

        # Update BalanceChart
        self.update_chart(balance_data)


class TransactionSimulator(QMainWindow):
    def __init__(self) -> None:
        super().__init__()

        self.transactions = [
            (1625692800, 0.1),  # July 8, 2021
            (1626111600, -0.03),  # July 13, 2021
            (1626520000, 0.15),  # July 18, 2021
            (1629138000, -0.05),  # August 17, 2021
            (1630547200, 0.08),  # September 3, 2021
        ]

        # Create BalanceChart
        self.chart = BalanceChart()
        self.setCentralWidget(self.chart)

        # QTimer to simulate incoming transactions
        self.timer = QTimer()
        self.timer.timeout.connect(self.add_transaction)
        self.timer.start(3000)

        # Initial chart update
        self.update_chart()

    def update_chart(self) -> None:
        # Calculate balance
        balance: float = 0
        balance_data = []
        for timestamp, amount in self.transactions:
            balance += amount
            balance_data.append((timestamp, balance))

        # Update BalanceChart
        self.chart.update_chart(balance_data)

    def add_transaction(self) -> None:
        # Simulating a new transaction (current timestamp, random amount)
        new_transaction = (
            int(
                (
                    datetime.datetime.fromtimestamp(self.transactions[-1][0]) + datetime.timedelta(days=5)
                ).timestamp()
            ),
            random.uniform(-0.1, 0.1),
        )
        self.transactions.append(new_transaction)

        # Update the chart
        self.update_chart()


if __name__ == "__main__":
    app = QApplication(sys.argv)
    main = TransactionSimulator()
    main.show()
    sys.exit(app.exec())



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import enum
import logging
from typing import Any, Callable, Dict, List, Optional, Set, Tuple

import bdkpython as bdk
from packaging import version
from PyQt6.QtCore import QObject

from .storage import BaseSaveableClass, SaveAllClass
from .util import Satoshis, serialized_to_hex

logger = logging.getLogger(__name__)


def is_address(a: str, network: bdk.Network) -> bool:
    try:
        bdk.Address(a, network=network)
    except:
        return False
    return True


class Recipient:
    def __init__(
        self, address: str, amount: int, label: Optional[str] = None, checked_max_amount=False
    ) -> None:
        self.address = address
        self.amount = amount
        self.label = label
        self.checked_max_amount = checked_max_amount

    def __hash__(self) -> int:
        "Necessary for the caching"
        return hash(self.__dict__)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__dict__})"

    def clone(self) -> "Recipient":
        return Recipient(self.address, self.amount, self.label, self.checked_max_amount)


class OutPoint(bdk.OutPoint):
    def __key__(self) -> Tuple:
        return tuple(v for k, v in sorted(self.__dict__.items()))

    def __hash__(self) -> int:
        "Necessary for the caching"
        return hash(self.__key__())

    def __str__(self) -> str:
        return f"{self.txid}:{self.vout}"

    def __repr__(self) -> str:
        return str(f"{self.__class__.__name__}({self.txid},{self.vout})")

    def __eq__(self, other) -> bool:
        if isinstance(other, OutPoint):
            return (self.txid, self.vout) == (other.txid, other.vout)
        return False

    @classmethod
    def from_bdk(cls, bdk_outpoint: bdk.OutPoint) -> "OutPoint":
        if isinstance(bdk_outpoint, OutPoint):
            return bdk_outpoint
        if isinstance(bdk_outpoint, str):
            return cls.from_str(bdk_outpoint)
        return OutPoint(bdk_outpoint.txid, bdk_outpoint.vout)

    @classmethod
    def from_str(cls, outpoint_str: str) -> "OutPoint":
        if isinstance(outpoint_str, OutPoint):
            return outpoint_str
        txid, vout = outpoint_str.split(":")
        return OutPoint(txid, int(vout))


def get_outpoints(tx: bdk.Transaction) -> List[OutPoint]:
    return [OutPoint.from_bdk(input.previous_output) for input in tx.input()]


class TxOut(bdk.TxOut):
    def __key__(self) -> Tuple:
        return (serialized_to_hex(self.script_pubkey.to_bytes()), self.value)

    def __hash__(self) -> int:
        "Necessary for the caching"
        return hash(self.__key__())

    def __str__(self) -> str:
        return str(self.__key__())

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__key__()})"

    @classmethod
    def from_bdk(cls, tx_out: bdk.TxOut) -> "TxOut":
        if isinstance(tx_out, TxOut):
            return tx_out
        return TxOut(tx_out.value, tx_out.script_pubkey)

    def __eq__(self, other) -> bool:
        if isinstance(other, TxOut):
            return (self.value, serialized_to_hex(self.script_pubkey.to_bytes())) == (
                other.value,
                serialized_to_hex(other.script_pubkey.to_bytes()),
            )
        return False


class PythonUtxo:
    def __init__(self, address: str, outpoint: OutPoint, txout: TxOut) -> None:
        self.outpoint = outpoint
        self.txout = txout
        self.address = address
        self.is_spent_by_txid: Optional[str] = None


def python_utxo_balance(python_utxos: List[PythonUtxo]) -> int:
    return sum([python_utxo.txout.value for python_utxo in python_utxos])


class UtxosForInputs:
    def __init__(
        self,
        utxos: List[PythonUtxo],
        included_opportunistic_merging_utxos=None,
        spend_all_utxos=False,
    ) -> None:
        if included_opportunistic_merging_utxos is None:
            included_opportunistic_merging_utxos = []

        self.utxos = utxos
        self.included_opportunistic_merging_utxos = included_opportunistic_merging_utxos
        self.spend_all_utxos = spend_all_utxos


class FullTxDetail:
    """For all outputs and inputs, where it has a full PythonUtxo ."""

    def __init__(self, tx: bdk.TransactionDetails, received=None, send=None) -> None:
        self.outputs: Dict[str, PythonUtxo] = received if received else {}  # outpoint_str: PythonUtxo
        self.inputs: Dict[str, Optional[PythonUtxo]] = send if send else {}  # outpoint_str: PythonUtxo
        self.tx = tx
        self.txid = tx.txid

    def involved_addresses(self) -> Set[str]:
        input_addresses = [input.address for input in self.inputs.values() if input]
        output_addresses = [output.address for output in self.outputs.values() if output]
        return set(input_addresses).union(output_addresses)

    @classmethod
    def fill_received(
        cls, tx: bdk.TransactionDetails, get_address_of_txout: Callable[[TxOut], str | None]
    ) -> "FullTxDetail":
        res = FullTxDetail(tx)
        txid = tx.txid
        for vout, txout in enumerate(tx.transaction.output()):
            address = get_address_of_txout(TxOut.from_bdk(txout))
            if not address:
                logger.error(
                    f"Could not calculate the address of {TxOut.from_bdk(txout)}. This should not happen, unless it is a mining input."
                )
                continue
            out_point = OutPoint(txid, vout)
            python_utxo = PythonUtxo(address, out_point, txout)
            python_utxo.is_spent_by_txid = None
            res.outputs[str(out_point)] = python_utxo
        return res

    def fill_inputs(
        self,
        lookup_dict_fulltxdetail: Dict[str, "FullTxDetail"],
    ) -> None:
        for prev_outpoint in get_outpoints(self.tx.transaction):
            prev_outpoint_str = str(prev_outpoint)

            # check if I have the prev_outpoint fulltxdetail
            if prev_outpoint.txid not in lookup_dict_fulltxdetail:
                self.inputs[prev_outpoint_str] = None
                continue
            fulltxdetail = lookup_dict_fulltxdetail[prev_outpoint.txid]
            if prev_outpoint_str not in fulltxdetail.outputs:
                self.inputs[prev_outpoint_str] = None
                continue
            python_utxo = fulltxdetail.outputs[prev_outpoint_str]
            python_utxo.is_spent_by_txid = self.tx.txid
            self.inputs[prev_outpoint_str] = python_utxo

    def sum_outputs(self, address_domain: List[str]) -> int:
        return sum(
            [
                python_utxo.txout.value
                for python_utxo in self.outputs.values()
                if python_utxo and python_utxo.address in address_domain
            ]
        )

    def sum_inputs(self, address_domain: List[str]) -> int:
        return sum(
            [
                python_utxo.txout.value
                for python_utxo in self.inputs.values()
                if python_utxo and python_utxo.address in address_domain
            ]
        )


class AddressInfoMin(SaveAllClass):
    def __init__(self, address: str, index: int, keychain: bdk.KeychainKind) -> None:
        self.address = address
        self.index = index
        self.keychain = keychain

    @classmethod
    def from_dump_migration(cls, dct: Dict[str, Any]) -> Dict[str, Any]:
        if version.parse(str(dct["VERSION"])) <= version.parse("0.0.0"):
            pass

        # now the version is newest, so it can be deleted from the dict
        if "VERSION" in dct:
            del dct["VERSION"]
        return dct

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__dict__})"

    def __key__(self) -> Tuple:
        return tuple(v for k, v in sorted(self.__dict__.items()))

    def __hash__(self) -> int:
        "Necessary for the caching"
        return hash(self.__key__())

    @classmethod
    def from_bdk_address_info(cls, bdk_address_info: bdk.AddressInfo) -> "AddressInfoMin":
        return AddressInfoMin(
            bdk_address_info.address.as_string(),
            bdk_address_info.index,
            bdk_address_info.keychain,
        )

    def is_change(self) -> bool:
        return self.keychain == bdk.KeychainKind.INTERNAL

    def address_path(self) -> Tuple[int, int]:
        return (bool(self.is_change()), self.index)

    @staticmethod
    def is_change_to_keychain(is_change: bool) -> bdk.KeychainKind:
        if is_change:
            return bdk.KeychainKind.INTERNAL
        else:
            return bdk.KeychainKind.EXTERNAL


class BlockchainType(enum.Enum):
    CompactBlockFilter = enum.auto()
    Electrum = enum.auto()
    Esplora = enum.auto()
    RPC = enum.auto()

    @classmethod
    def from_text(cls, t) -> "BlockchainType":
        if t == "Compact Block Filters":
            return cls.CompactBlockFilter
        elif t == "Electrum Server":
            return cls.Electrum
        elif t == "Esplora Server":
            return cls.Esplora
        elif t == "RPC":
            return cls.RPC

        raise ValueError(f"{t} is not a valid BlockchainType")

    @classmethod
    def to_text(cls, t) -> str:
        if t == cls.CompactBlockFilter:
            return "Compact Block Filters"
        elif t == cls.Electrum:
            return "Electrum Server"
        elif t == cls.Esplora:
            return "Esplora Server"
        elif t == cls.RPC:
            return "RPC"
        else:
            raise ValueError()

    @classmethod
    def active_types(cls) -> List["BlockchainType"]:
        return [cls.Electrum, cls.Esplora, cls.RPC]


class CBFServerType(enum.Enum):
    Automatic = enum.auto()
    Manual = enum.auto()

    @classmethod
    def from_text(cls, t) -> "CBFServerType":
        return CBFServerType._member_map_[t]  # type: ignore


class Balance(QObject, SaveAllClass):
    VERSION = "0.0.1"
    known_classes = {
        **BaseSaveableClass.known_classes,
    }

    def __init__(self, immature=0, trusted_pending=0, untrusted_pending=0, confirmed=0) -> None:
        super().__init__()
        self.immature = immature
        self.trusted_pending = trusted_pending
        self.untrusted_pending = untrusted_pending
        self.confirmed = confirmed

    @property
    def total(self) -> int:
        return self.immature + self.trusted_pending + self.untrusted_pending + self.confirmed

    @property
    def spendable(self) -> int:
        return self.trusted_pending + self.confirmed

    def __add__(self, other: "Balance") -> "Balance":
        summed = {key: self.__dict__[key] + other.__dict__[key] for key in self.__dict__.keys()}
        return self.__class__(**summed)

    def format_long(self, network: bdk.Network) -> str:

        details = [
            f"{title}: {Satoshis (value, network=network).str_with_unit()}"
            for title, value in [
                (self.tr("Confirmed"), self.confirmed),
                (
                    self.tr("Unconfirmed"),
                    self.untrusted_pending + self.trusted_pending,
                ),
                (self.tr("Unmatured"), self.immature),
            ]
        ]
        long = "\n".join(details)
        return long

    def format_short(self, network: bdk.Network) -> str:

        short = Satoshis(value=self.total, network=network).format_as_balance()

        return short

    @classmethod
    def from_dump_migration(cls, dct: Dict[str, Any]) -> Dict[str, Any]:
        if version.parse(str(dct["VERSION"])) <= version.parse("0.0.0"):
            pass

        # now the version is newest, so it can be deleted from the dict
        if "VERSION" in dct:
            del dct["VERSION"]
        return dct


def robust_address_str_from_script(script_pubkey: bdk.Script, network, on_error_return_hex=True) -> str:
    try:
        return bdk.Address.from_script(script_pubkey, network).as_string()
    except:
        if on_error_return_hex:
            return serialized_to_hex(script_pubkey.to_bytes())
        else:
            return ""


if __name__ == "__main__":
    testdict = {}

    def test_hashing(v) -> None:
        testdict[v] = v.__hash__()
        print(testdict[v])

    test_hashing(OutPoint("txid", 0))
    test_hashing(AddressInfoMin("ssss", 4, bdk.KeychainKind.EXTERNAL))



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


"""
This file was created once to provide  the pubkeys of a TxIn

- since it wasnt needed anywhere the functions were removed again
"""
import logging

logger = logging.getLogger(__name__)


import bdkpython as bdk

from bitcoin_safe.util import hex_to_serialized


def script_pubkey_to_address(script_pubkey: str, network: bdk.Network) -> str:
    return bdk.Address.from_script(bdk.Script(hex_to_serialized(script_pubkey)), network).as_string()



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging

from bitcoin_qr_tools.multipath_descriptor import (
    MultipathDescriptor as BitcoinQRMultipathDescriptor,
)

from bitcoin_safe.gui.qt.descriptor_edit import DescriptorEdit
from bitcoin_safe.gui.qt.dialogs import question_dialog
from bitcoin_safe.gui.qt.keystore_uis import KeyStoreUIs
from bitcoin_safe.i18n import translate
from bitcoin_safe.threading_manager import ThreadingManager

logger = logging.getLogger(__name__)

from typing import Callable, Optional, Tuple

from bitcoin_usb.address_types import get_address_types
from PyQt6.QtCore import QMargins, QObject, Qt, pyqtSignal
from PyQt6.QtWidgets import (
    QComboBox,
    QDialogButtonBox,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QSizePolicy,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from ...descriptors import AddressType, get_default_address_type
from ...signals import SignalsMin, pyqtSignal
from ...wallet import ProtoWallet, Wallet
from .block_change_signals import BlockChangesSignals


class DescriptorUI(QObject):
    signal_qtwallet_apply_setting_changes = pyqtSignal()
    signal_qtwallet_cancel_setting_changes = pyqtSignal()
    signal_qtwallet_cancel_wallet_creation = pyqtSignal()

    def __init__(
        self,
        protowallet: ProtoWallet,
        signals_min: SignalsMin,
        get_lang_code: Callable[[], str],
        get_wallet: Optional[Callable[[], Wallet]] = None,
        threading_parent: ThreadingManager | None = None,
    ) -> None:
        super().__init__()
        # if we are in the wallet setp process, then wallet = None
        self.protowallet = protowallet
        self.get_wallet = get_wallet
        self.signals_min = signals_min
        self.get_lang_code = get_lang_code
        self.threading_parent = threading_parent

        self.no_edit_mode = (self.protowallet.threshold, len(self.protowallet.keystores)) in [(1, 1), (2, 3)]

        self.tab = QWidget()
        self.tab_layout = QVBoxLayout(self.tab)

        self.create_wallet_type_and_descriptor()

        self.repopulate_comboBox_address_type(self.protowallet.is_multisig())

        self.edit_descriptor.signal_descriptor_change.connect(self.on_descriptor_change)

        self.keystore_uis = KeyStoreUIs(
            get_editable_protowallet=lambda: self.protowallet,
            get_address_type=self.get_address_type_from_ui,
            signals_min=signals_min,
        )
        self.tab_layout.addWidget(self.keystore_uis)

        self.keystore_uis.setCurrentIndex(0)

        self.set_all_ui_from_protowallet()
        # diasbeling fields MUST be done after the ui is filled
        self.disable_fields()

        self.box_button_bar = self.create_button_bar()
        self.updateUi()
        signals_min.language_switch.connect(self.updateUi)

    def updateUi(self) -> None:
        self.label_signers.setText(self.tr("Required Signers"))
        self.label_gap.setText(self.tr("Scan Addresses ahead"))
        self.edit_descriptor.input_field.setPlaceholderText(
            self.tr("Paste or scan your descriptor, if you restore a wallet.")
        )

        self.edit_descriptor.setToolTip(
            self.tr(
                'This "descriptor" contains all information to reconstruct the wallet. \nPlease back up this descriptor to be able to recover the funds!'
            )
        )
        self.box_wallet_type.setTitle(translate("descriptor", "Wallet Properties"))
        self.label_address_type.setText(translate("descriptor", "Address Type"))
        self.groupBox_wallet_descriptor.setTitle(translate("descriptor", "Wallet Descriptor"))

    def set_protowallet(self, protowallet: ProtoWallet) -> None:
        self.protowallet = protowallet
        self.set_all_ui_from_protowallet()

    def on_wallet_ui_changes(self) -> None:
        logger.debug("on_wallet_ui_changes")
        try:
            self.set_protowallet_from_ui()

            self.set_ui_descriptor()
            self.keystore_uis.set_keystore_ui_from_protowallet()
            self.set_wallet_ui_from_protowallet()
        except:
            logger.warning("on_wallet_ui_changes: Invalid input")
            self._set_keystore_tabs()

    def on_descriptor_change(self, new_value: str) -> None:
        if not BitcoinQRMultipathDescriptor.is_valid(new_value, network=self.protowallet.network):
            logger.debug("Descriptor invalid")
            return

        old_descriptor = self.protowallet.to_multipath_descriptor()

        if old_descriptor and (new_value == old_descriptor.as_string()):
            logger.info("Descriptor unchanged")
            return
        else:
            logger.info(f"Descriptor changed: {old_descriptor}  -->  {new_value}")
            if not question_dialog(
                text=self.tr(
                    f"Fill signer information based on the new descriptor?",
                ),
                title=self.tr("New descriptor entered"),
                buttons=QMessageBox.StandardButton.No | QMessageBox.StandardButton.Yes,
            ):
                return

        try:
            self.set_protowallet_from_descriptor_str(new_value)
            logger.info(f"Successfully set protwallet from descriptor {new_value}")

            self.set_wallet_ui_from_protowallet()
            self.keystore_uis.set_keystore_ui_from_protowallet()
            self.disable_fields()
        except:
            logger.info(f"Invalid descriptor {new_value}")
            return

    def on_spin_threshold_changed(self, new_value: int) -> None:
        self.on_wallet_ui_changes()

    def on_spin_signer_changed(self, new_value: int) -> None:
        self.repopulate_comboBox_address_type(new_value > 1)

        self.on_wallet_ui_changes()

    def set_protowallet_from_descriptor_str(self, descriptor: str) -> None:
        self.protowallet = ProtoWallet.from_descriptor(
            self.protowallet.id, descriptor, self.protowallet.network
        )

    def _set_keystore_tabs(self) -> None:
        self.keystore_uis._set_keystore_tabs()

        self.spin_req.setMinimum(1)
        self.spin_req.setMaximum(self.spin_signers.value())
        self.spin_signers.setMinimum(self.spin_req.value())
        self.spin_signers.setMaximum(10)

    def set_wallet_ui_from_protowallet(self) -> None:
        with BlockChangesSignals([self.tab]):
            logger.debug(f"{self.__class__.__name__} set_wallet_ui_from_protowallet")
            self.repopulate_comboBox_address_type(self.protowallet.is_multisig())
            self.comboBox_address_type.setCurrentText(self.protowallet.address_type.name)
            self.spin_req.setMinimum(1)
            self.spin_req.setMaximum(len(self.protowallet.keystores))
            self.spin_req.setValue(self.protowallet.threshold)

            self.spin_signers.setMinimum(self.protowallet.threshold)
            self.spin_signers.setMaximum(10)
            self.spin_signers.setValue(len(self.protowallet.keystores))

            if self.spin_req.value() < self.spin_signers.value():

                labels_of_recovery_signers = [
                    f'"{keystore_ui.label}"' for keystore_ui in self.keystore_uis.getAllTabData().values()
                ][self.spin_req.value() :]
                self.spin_req.setToolTip(
                    f"In the chosen multisig setup, you need {self.spin_req.value()} devices (signers) to sign every outgoing transaction.\n"
                    f'In case of loss of 1 of the devices, you can recover your funds using\n {" or ".join(labels_of_recovery_signers)} and send the funds to a new wallet.'
                )
            if self.spin_req.value() == self.spin_signers.value() != 1:
                self.spin_req.setToolTip(
                    f"Warning!  Choosing a multisig setup where ALL signers need to sign every transaction\n is very RISKY and does not offer any benefits of multisig. Recommended multisig setups are 2-of-3 or 3-of-5"
                )
            if self.spin_req.value() == self.spin_signers.value() == 1:
                self.spin_req.setToolTip(f"A single signing device can sign outgoing transactions.")

            self.spin_gap.setValue(self.protowallet.gap)

    def set_all_ui_from_protowallet(self) -> None:
        """Updates the 3 parts.

        - wallet ui (e.g. gap)
        - Keystore UI  (e.g. xpubs)
        - descriptor ui
        """
        with BlockChangesSignals([self.tab]):
            self.keystore_uis.set_keystore_ui_from_protowallet()
            self.set_wallet_ui_from_protowallet()
            self.set_ui_descriptor()

    def set_protowallet_from_ui(self) -> None:
        logger.debug("set_protowallet_from_keystore_ui")

        # these wallet settings must come first
        m, n = self.get_m_of_n_from_ui()
        self.protowallet.set_number_of_keystores(n)
        self.protowallet.set_threshold(m)
        self.protowallet.set_address_type(self.get_address_type_from_ui())
        self.protowallet.set_gap(self.get_gap_from_ui())

        self.keystore_uis.set_protowallet_from_keystore_ui()

    def set_combo_box_address_type_default(self) -> None:
        address_types = get_address_types(self.protowallet.is_multisig())
        self.comboBox_address_type.setCurrentIndex(
            address_types.index(get_default_address_type(self.protowallet.is_multisig()))
        )

    def get_address_type_from_ui(self) -> AddressType:
        address_types = get_address_types(self.protowallet.is_multisig())

        # sanity check
        assert (
            self.comboBox_address_type.currentText()
            == address_types[self.comboBox_address_type.currentIndex()].name
        )
        assert (
            self.comboBox_address_type.currentData()
            == address_types[self.comboBox_address_type.currentIndex()]
        )

        return self.comboBox_address_type.currentData()

    def get_m_of_n_from_ui(self) -> Tuple[int, int]:
        return (self.spin_req.value(), self.spin_signers.value())

    def get_gap_from_ui(self) -> int:
        return self.spin_gap.value()

    def set_ui_descriptor(self) -> None:
        logger.debug(f"{self.__class__.__name__} set_ui_descriptor")
        # check if the descriptor actually CAN be calculated to a reasonable degree
        try:
            multipath_descriptor = self.protowallet.to_multipath_descriptor()
            if multipath_descriptor:
                self.edit_descriptor.setText(multipath_descriptor.as_string_private())
            else:
                self.edit_descriptor.setText("")
        except:
            self.edit_descriptor.setText("")
        self.edit_descriptor.format_and_apply_validator()

    def disable_fields(self) -> None:
        self.comboBox_address_type.setHidden(self.no_edit_mode)
        self.label_address_type.setHidden(self.no_edit_mode)
        self.spin_signers.setHidden(self.no_edit_mode)
        self.spin_req.setHidden(self.no_edit_mode)
        self.label_signers.setHidden(self.no_edit_mode)
        self.label_of.setHidden(self.no_edit_mode)

        with BlockChangesSignals([self.tab]):
            self.set_combo_box_address_type_default()
            self.spin_signers.setValue(len(self.protowallet.keystores))

        if self.protowallet.is_multisig():
            self.label_of.setEnabled(True)
            self.spin_signers.setEnabled(True)
        else:
            self.label_of.setDisabled(True)
            self.spin_signers.setDisabled(True)

    def repopulate_comboBox_address_type(self, is_multisig: bool) -> None:
        with BlockChangesSignals([self.tab]):
            # Fetch the new address types
            address_types = get_address_types(is_multisig)
            address_type_names = [a.name for a in address_types]

            # Get the current items in the combo box
            current_names = [
                self.comboBox_address_type.itemText(i) for i in range(self.comboBox_address_type.count())
            ]

            # Check if the new list is different from the current items
            if address_type_names != current_names:

                # Clear and update the combo box
                self.comboBox_address_type.clear()
                for address_type in address_types:
                    self.comboBox_address_type.addItem(address_type.name, userData=address_type)

                default_address_type = get_default_address_type(is_multisig).name
                if default_address_type in address_type_names:
                    self.comboBox_address_type.setCurrentIndex(address_type_names.index(default_address_type))

    def create_wallet_type_and_descriptor(self) -> None:
        box_wallet_type_and_descriptor = QWidget(self.tab)
        box_wallet_type_and_descriptor_layout = QHBoxLayout(box_wallet_type_and_descriptor)

        current_margins = box_wallet_type_and_descriptor_layout.contentsMargins()
        box_wallet_type_and_descriptor_layout.setContentsMargins(
            QMargins(0, 0, 0, current_margins.bottom())
        )  # Smaller margins (left, top, right, bottom)

        # Removed the unnecessary parent widgets. Using QGroupBox directly as the container.
        self.box_wallet_type = QGroupBox()

        # Create a QFormLayout
        form_wallet_type = QGridLayout(self.box_wallet_type)

        # box_signers_with_slider
        self.label_signers = QLabel()

        self.spin_req = QSpinBox()
        self.spin_req.setObjectName("spin_req")
        self.spin_req.setMinimum(1)
        self.spin_req.setMaximum(10)

        self.label_of = QLabel()
        self.label_of.setText("of")
        self.label_of.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.spin_signers = QSpinBox()
        self.spin_signers.setObjectName("spin_signers")
        self.spin_signers.setMinimum(1)
        self.spin_signers.setMaximum(10)

        # Add widgets to the layout
        form_wallet_type.addWidget(self.label_signers, 0, 0)
        form_wallet_type.addWidget(self.spin_req, 0, 1)
        form_wallet_type.addWidget(self.label_of, 0, 2)
        form_wallet_type.addWidget(self.spin_signers, 0, 3)

        # box_address_type
        self.label_address_type = QLabel()
        self.label_address_type.setObjectName("this label")

        self.comboBox_address_type = QComboBox()
        self.comboBox_address_type.setObjectName("this QComboBox")
        self.comboBox_address_type.currentIndexChanged.connect(self.on_wallet_ui_changes)
        form_wallet_type.addWidget(self.label_address_type, 2, 0)
        form_wallet_type.setObjectName("this form_wallet_type")
        form_wallet_type.addWidget(self.comboBox_address_type, 2, 1, 1, 3)

        # box_gap
        self.label_gap = QLabel()
        self.label_gap.setWordWrap(True)

        self.spin_gap = QSpinBox()
        self.spin_gap.setMinimum(20)
        self.spin_gap.setMaximum(int(1e6))

        # Add widgets to the layout
        form_wallet_type.addWidget(self.label_gap, 3, 0)
        form_wallet_type.addWidget(self.spin_gap, 3, 1, 1, 3)

        self.box_wallet_type.setLayout(form_wallet_type)
        box_wallet_type_and_descriptor_layout.addWidget(self.box_wallet_type)

        # now the descriptor
        self.groupBox_wallet_descriptor = QGroupBox()
        self.groupBox_wallet_descriptor.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum
        )
        # below is an example how to highlight the box
        # groupBox_wallet_descriptor.setStyleSheet("""
        # QGroupBox {
        #         font-weight: bold;
        #         border: 2px solid red;
        #         border-radius: 5px;
        #         margin-top: 12px;
        # }
        # QGroupBox::title {
        #         color: red;
        #         subcontrol-origin: margin;
        #         left: 10px;
        #         padding: 0 5px 0 5px;
        # }
        # """)
        self.horizontalLayout_4 = QHBoxLayout(self.groupBox_wallet_descriptor)
        self.edit_descriptor = DescriptorEdit(
            network=self.protowallet.network,
            signals_min=self.signals_min,
            get_wallet=self.get_wallet,
            signal_update=self.signals_min.language_switch,
            threading_parent=self.threading_parent,
            get_lang_code=self.get_lang_code,
        )
        self.edit_descriptor.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_4.addWidget(self.edit_descriptor)

        box_wallet_type_and_descriptor_layout.addWidget(self.groupBox_wallet_descriptor)

        self.tab_layout.addWidget(box_wallet_type_and_descriptor)

        self.spin_signers.valueChanged.connect(self.on_spin_signer_changed)
        self.spin_req.valueChanged.connect(self.on_spin_threshold_changed)

    def create_button_bar(self) -> QDialogButtonBox:

        # Create buttons and layout
        self.button_box = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Apply | QDialogButtonBox.StandardButton.Discard
        )
        if _button := self.button_box.button(QDialogButtonBox.StandardButton.Apply):
            _button.clicked.connect(self.signal_qtwallet_apply_setting_changes.emit)
        if _button := self.button_box.button(QDialogButtonBox.StandardButton.Discard):
            _button.clicked.connect(self.signal_qtwallet_cancel_setting_changes.emit)
        if _button := self.button_box.button(QDialogButtonBox.StandardButton.Discard):
            _button.clicked.connect(self.signal_qtwallet_cancel_wallet_creation.emit)

        self.tab_layout.addWidget(self.button_box, 0, Qt.AlignmentFlag.AlignRight)
        return self.button_box



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import json

import requests


def send_rpc_command(ip: str, port: str, username: str, password: str, method: str, params=None) -> str:
    """Sends an RPC command to a Bitcoin node.

    :param ip: IP address of the Bitcoin node.
    :param port: RPC port of the Bitcoin node.
    :param username: RPC username.
    :param password: RPC password.
    :param method: RPC method/command to execute.
    :param params: Parameters for the RPC method (default: empty list).
    :return: The response of the RPC command.
    """
    if not params:
        params = []

    # Create the URL for the RPC endpoint
    url = f"http://{ip}:{port}"

    # Create the headers
    headers = {"content-type": "application/json"}

    # Create the payload with the RPC command and parameters
    payload = json.dumps(
        {
            "method": method,
            "params": params,
            "id": "1",  # This can be any ID, used for identifying the request
        }
    )

    # Send the request and get the response
    response = requests.post(url, headers=headers, data=payload, auth=(username, password), timeout=20)

    # Return the response
    return response.json()



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from typing import Dict, List, Optional

import bdkpython as bdk
from PyQt6.QtGui import QColor

from bitcoin_safe.gui.qt.address_edit import AddressEdit
from bitcoin_safe.gui.qt.sankey_widget import FlowIndex, FlowType, SankeyWidget
from bitcoin_safe.html import html_f
from bitcoin_safe.psbt_util import FeeInfo
from bitcoin_safe.pythonbdk_types import (
    OutPoint,
    PythonUtxo,
    TxOut,
    get_outpoints,
    robust_address_str_from_script,
)
from bitcoin_safe.signals import Signals, UpdateFilter
from bitcoin_safe.util import Satoshis
from bitcoin_safe.wallet import Wallet, get_label_from_any_wallet, get_wallets

logger = logging.getLogger(__name__)


class SankeyBitcoin(SankeyWidget):
    def __init__(self, network: bdk.Network, signals: Signals):
        super().__init__()
        self.signals = signals
        self.network = network
        self.tx: bdk.Transaction | None = None
        self.txouts: List[TxOut] = []
        self.addresses: List[str] = []

        self.signals.any_wallet_updated.connect(self.refresh)
        self.signal_on_label_click.connect(self.on_label_click)

    def refresh(self, update_filter: UpdateFilter):
        if not self.tx:
            return

        should_update = False
        if should_update or update_filter.refresh_all:
            should_update = True
        if should_update or self.tx.txid() in update_filter.txids:
            should_update = True
        if should_update or set(self.outpoints).intersection(update_filter.outpoints):
            should_update = True
        if should_update or set(self.input_outpoints).intersection(update_filter.outpoints):
            should_update = True
        if should_update or set(self.addresses).intersection(update_filter.addresses):
            should_update = True

        if not should_update:
            return

        logger.debug(f"{self.__class__.__name__} update_with_filter {update_filter}")
        self.set_tx(self.tx)

    @property
    def outpoints(self) -> List[OutPoint]:
        if not self.tx:
            return []
        txid = self.tx.txid()
        return [OutPoint(txid=txid, vout=vout) for vout in range(len(self.tx.output()))]

    @property
    def input_outpoints(self) -> List[OutPoint]:
        if not self.tx:
            return []
        return [OutPoint.from_bdk(inp.previous_output) for inp in self.tx.input()]

    def set_tx(self, tx: bdk.Transaction, fee_info: FeeInfo | None = None) -> bool:
        self.tx = tx
        self.addresses = []
        wallets = get_wallets(self.signals)

        labels: Dict[FlowIndex, str] = {}
        tooltips: Dict[FlowIndex, str] = {}
        colors: Dict[FlowIndex, QColor] = {}

        # output
        self.txouts = [TxOut.from_bdk(txout) for txout in tx.output()]
        out_flows: List[float] = [txout.value for txout in self.txouts]
        for i, txout in enumerate(self.txouts):
            flow_index = FlowIndex(flow_type=FlowType.OutFlow, i=i)
            address = robust_address_str_from_script(txout.script_pubkey, network=self.network)
            self.addresses.append(address)

            label = get_label_from_any_wallet(
                address, signals=self.signals, wallets=wallets, autofill_from_txs=False
            )
            color = self.get_address_color(address, wallets=wallets)
            labels[flow_index] = label if label else address
            tooltips[flow_index] = html_f(
                ((label + "\n" + address) if label else address)
                + "\n"
                + Satoshis(txout.value, self.network).str_with_unit(),
                add_html_and_body=True,
            )
            if color:
                colors[flow_index] = color

        wallets = get_wallets(self.signals)
        outpoint_dict = {
            outpoint_str: (python_utxo, wallet)
            for wallet in wallets
            for outpoint_str, python_utxo in wallet.get_all_txos_dict().items()
        }

        # input
        in_python_txos: List[PythonUtxo] = []
        sufficient_info = True
        for outpoint in get_outpoints(tx):
            outpoint_str = str(outpoint)
            if outpoint_str not in outpoint_dict:
                # ensure all inputs are known
                sufficient_info = False
                break
            python_utxo, wallet = outpoint_dict[outpoint_str]
            in_python_txos.append(python_utxo)

        for i, txo in enumerate(in_python_txos):
            self.addresses.append(txo.address)
            flow_index = FlowIndex(flow_type=FlowType.InFlow, i=i)

            label = get_label_from_any_wallet(
                txo.address, signals=self.signals, wallets=wallets, autofill_from_txs=False
            )
            color = self.get_address_color(txo.address, wallets=wallets)
            labels[flow_index] = label if label else txo.address
            tooltips[flow_index] = html_f(
                ((label + "\n" + txo.address) if label else txo.address)
                + "\n"
                + Satoshis(txo.txout.value, self.network).str_with_unit(),
                add_html_and_body=True,
            )
            if color:
                colors[flow_index] = color

        in_flows: List[float] = [txo.txout.value for txo in in_python_txos]

        if not sufficient_info:
            # if there is only 1 input and the fee is known, I can still construct a diagram
            if len(get_outpoints(tx)) == 1 and len(in_flows) == 0 and fee_info and not fee_info.is_estimated:
                in_flows = [sum(out_flows) + fee_info.fee_amount]
                sufficient_info = True

        if not sufficient_info:
            return False

        in_sum = sum(in_flows)

        # other
        fee = int(in_sum - sum(out_flows))
        if fee > 0:
            out_flows.append(fee)
            flow_index = FlowIndex(FlowType.OutFlow, i=len(self.txouts))
            labels[flow_index] = self.tr("Fee")
            tooltips[flow_index] = html_f(
                labels[flow_index] + "<br>" + Satoshis(fee, self.network).str_with_unit(),
                add_html_and_body=True,
            )

        self.set(
            in_flows=in_flows,
            out_flows=out_flows,
            colors=colors,
            labels=labels,
            tooltips=tooltips,
        )
        return True

    def get_address_color(self, address: str, wallets: List[Wallet]) -> QColor | None:
        def get_wallet():
            for wallet in wallets:
                if wallet.is_my_address(address):
                    return wallet
            return None

        wallet = get_wallet()
        if not wallet:
            return None
        color = AddressEdit.color_address(address, wallet)
        if not color:
            logger.error("This should not happen, since wallet should only be found if the address is mine.")
            return None
        return color

    def get_python_txo(self, outpoint: str, wallets: List[Wallet] | None = None) -> Optional[PythonUtxo]:
        wallets = wallets if wallets else get_wallets(self.signals)
        for wallet in wallets:
            txo = wallet.get_python_txo(outpoint)
            if txo:
                return txo
        return None

    def on_label_click(self, flow_index: FlowIndex):
        if not self.tx:
            return
        if flow_index.flow_type == FlowType.OutFlow:
            # output
            # careful, the last flow_index.i is the fee, so
            # outflow indexes go 1 larger than the actual vout index
            outpoint = OutPoint(self.tx.txid(), flow_index.i)
            txo = self.get_python_txo(str(outpoint))
            if not txo:
                return
            if txo.is_spent_by_txid:
                # open the spending tx
                self.signals.open_tx_like.emit(txo.is_spent_by_txid)

        elif flow_index.flow_type == FlowType.InFlow:
            outpoints = get_outpoints(self.tx)
            if len(outpoints) <= flow_index.i:
                return
            outpoint = outpoints[flow_index.i]
            self.signals.open_tx_like.emit(outpoint.txid)



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import random
from typing import Type, TypeVar

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QColor, QFont, QPainter, QPaintEvent
from PyQt6.QtWidgets import (
    QApplication,
    QPushButton,
    QStyleOption,
    QVBoxLayout,
    QWidget,
)


class DebugWidget(QWidget):
    def paintEvent(self, event: QPaintEvent | None) -> None:
        super().paintEvent(event)
        self.drawDebugInfo(self)

    def _cleaned_size_policy(self, policy) -> str:
        return str(policy).split(".")[-1]

    def _collect_debug_info(self, widget: QWidget, level=0) -> str:
        indent = "    " * level
        sizePolicy = widget.sizePolicy()
        sizePolicyText = f"{indent}SP: H-{self._cleaned_size_policy(sizePolicy.horizontalPolicy())}, V-{self._cleaned_size_policy(sizePolicy.verticalPolicy())}"
        # sizeText = f"{indent}Size: Min {widget.minimumWidth()}x{widget.minimumHeight()}, Max {widget.maximumWidth()}x{widget.maximumHeight()}"
        sizeText = f"{indent}sizeHint: {widget.sizeHint()}"
        classNameText = f"{indent}Class: {widget.__class__.__name__}"

        # Margins (for QLayout if exists)
        layout = widget.layout()
        if layout:
            margins = layout.getContentsMargins()
            marginText = f"{indent}Margins: {margins}"
        else:
            marginText = f"{indent}No layout/margins"

        # Padding (generalized)
        option = QStyleOption()
        option.initFrom(widget)
        padding = option.rect
        paddingText = f"{indent}Padding: {padding}"

        tooltipText = f"{classNameText}\n{sizePolicyText}\n{sizeText}\n{marginText}\n{paddingText}"

        # Recursively collect info from children
        for child in widget.children():
            if isinstance(child, QWidget) and level < 2:
                tooltipText += "\n\n" + self._collect_debug_info(child, level + 1)

        return tooltipText

    def drawDebugInfo(self, widget: QWidget) -> None:
        widget_hash = hash(widget)
        random.seed(widget_hash)
        color = QColor(random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))

        painter = QPainter(widget)
        painter.setPen(color)
        painter.drawRect(widget.rect().adjusted(0, 0, -1, -1))

        # Set tooltip with recursive information
        widget.setToolTip(self._collect_debug_info(widget))

        # Display size policy and size information on widget
        font = QFont("Arial", 8)
        painter.setFont(font)
        painter.drawText(
            widget.rect().adjusted(5, 5, -5, -5),
            Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignLeft,
            widget.__class__.__name__,
        )


W = TypeVar("W", bound=QWidget)


def generate_debug_class(BaseClass: Type[W]) -> Type[W]:
    class DebugClass(BaseClass):  # type: ignore
        def paintEvent(self, event: QPaintEvent) -> None:
            super().paintEvent(event)
            DebugWidget().drawDebugInfo(self)

    DebugClass.__name__ = f"{BaseClass.__name__}"
    return DebugClass  # type: ignore


if __name__ == "__main__":
    app = QApplication([])

    DebugButton = generate_debug_class(QPushButton)
    main_widget = DebugWidget()
    layout = QVBoxLayout(main_widget)
    layout.addWidget(DebugButton("Button 1"))
    layout.addWidget(DebugButton("Button 2"))

    main_widget.show()
    app.exec()



import random

from PyQt6.QtWidgets import QApplication

from .quick_receive import QuickReceive, ReceiveGroup


def generate_random_hex_color() -> str:
    """Generate a random hex color code."""
    random_color = "#{:06x}".format(random.randint(0, 0xFFFFFF))
    return random_color


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)

    custom_widget = QuickReceive("Quick Receive")

    # Example usage
    color = generate_random_hex_color()
    custom_widget.add_box(ReceiveGroup(color, generate_random_hex_color(), color * 10, color * 10))
    color = generate_random_hex_color()
    custom_widget.add_box(ReceiveGroup(color, generate_random_hex_color(), color * 10, color * 10))
    color = generate_random_hex_color()
    custom_widget.add_box(ReceiveGroup(color, generate_random_hex_color(), color * 10, color * 10))

    custom_widget.show()
    sys.exit(app.exec())



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging

logger = logging.getLogger(__name__)


import json
from dataclasses import dataclass
from math import ceil
from typing import Any, Dict, List, Tuple

import bdkpython as bdk

from .dynamic_lib_load import setup_libsecp256k1

setup_libsecp256k1()


from bitcoin_usb.address_types import SimplePubKeyProvider

from bitcoin_safe.util import remove_duplicates_keep_order

from .pythonbdk_types import OutPoint, TxOut


def parse_redeem_script(script_hex: str) -> Tuple[int, List[str]]:
    """
    Parses a redeem script to extract the threshold (m) and public keys.

    Parameters:
    - script_hex (str): The hexadecimal string of the redeem script.

    Returns:
    - Tuple[int, List[str]]: A tuple containing the threshold (m) and a list of public keys.
    """
    m = int(script_hex[:2], 16) - 80
    script_body = script_hex[2:]
    public_keys = []
    while script_body:
        if script_body[:2] == "21":
            key = script_body[2:68]
            public_keys.append(key)
            script_body = script_body[68:]
        else:
            break

    return m, public_keys


def parse_witness_script(script_hex: str) -> Tuple[int, List[str]]:
    """
    Parses a witness script to extract the threshold (m) and public keys.

    Parameters:
    - script_hex (str): The hexadecimal string of the witness script.

    Returns:
    - Tuple[int, List[str]]: A tuple containing the threshold (m) and a list of public keys.
    """
    m = int(script_hex[:2], 16) - 80
    script_body = script_hex[2:]
    public_keys = []
    while script_body:
        if script_body[:2] == "21":
            key = script_body[2:68]
            public_keys.append(key)
            script_body = script_body[68:]
        else:
            break

    return m, public_keys


def weight_to_vsize(weight) -> int:
    return ceil(weight / 4)


def estimate_tx_weight(
    input_mn_tuples: List[Tuple[int, int]], num_outputs: int, include_signatures=True
) -> int:
    """Estimate the weight of a SegWit transaction in weight units, including
    support for multiple inputs with varying m-of-n multisignature
    configurations.

    Args:
    input_mn_tuples (list of tuples): A list where each tuple represents an input with (m, n) configuration.
    num_outputs (int): Number of outputs in the transaction.
    include_signatures (bool): Whether to include the size of signatures in the estimate.

    Returns:
    int: Estimated transaction weight in weight units.
    """
    # Transaction overheads
    version_size = 4
    segwit_marker_and_flag_size = 2
    locktime_size = 4

    # Input components
    outpoint_size = 36  # txid (32 bytes) + vout index (4 bytes)
    script_length_size = 1  # Size of the script length field
    sequence_size = 4  # Size of the sequence field
    base_input_size = outpoint_size + sequence_size  # Excluding scriptSig and witness

    # Output components
    output_value_size = 8  # Size of the value field
    output_script_length_size = 1  # Size of the script length field
    p2wpkh_script_size = 22  # P2WPKH script size
    base_output_size = output_value_size + output_script_length_size + p2wpkh_script_size

    # Witness components
    witness_stack_items_size = 1  # Size byte for the number of witness stack items
    average_signature_size = 72  # Approximate size of a signature
    average_pubkey_size = 33  # Size of a compressed public key

    # Calculate total witness data size
    total_witness_data_size = 0
    for m, n in input_mn_tuples:
        witness_data_size_per_input = (
            witness_stack_items_size + (m * (1 + average_signature_size)) + (n * (1 + average_pubkey_size))
        )
        total_witness_data_size += witness_data_size_per_input if include_signatures else 0

    # Calculate base transaction size (excluding witness data)
    num_inputs = len(input_mn_tuples)
    base_tx_size_without_witness = (
        version_size
        + segwit_marker_and_flag_size
        + locktime_size
        + num_inputs * (base_input_size + script_length_size)
        + num_outputs * base_output_size
    )

    # Calculate transaction weight
    # Non-witness data is weighted as 4 units per byte, witness data as 1 unit per byte
    tx_weight: int = (base_tx_size_without_witness * 4) + total_witness_data_size

    return tx_weight


class FeeInfo:
    def __init__(self, fee_amount: int, vsize: int, is_estimated=False) -> None:
        """_summary_

        Args:
            fee_amount (int): _description_
            vsize (int): transaction.vsize()
            is_estimated (bool, optional): _description_. Defaults to False.
        """
        self.fee_amount = fee_amount
        self.vsize = vsize
        self.is_estimated = is_estimated

    def fee_rate(self) -> float:
        return self.fee_amount / self.vsize

    @classmethod
    def from_txdetails(cls, tx_details: bdk.TransactionDetails) -> "FeeInfo":
        fee = tx_details.fee
        # coinbase transaction have fee=None
        fee = fee if fee is not None else 0
        return FeeInfo(fee, tx_details.transaction.vsize(), is_estimated=False)

    @classmethod
    def estimate_segwit_fee_rate_from_psbt(cls, psbt: bdk.PartiallySignedTransaction) -> "FeeInfo":
        """Estimate the fee rate of a SegWit transaction from a serialized PSBT
        JSON.

        Args:
        psbt_json_str (str): The serialized PSBT JSON string.

        Returns:
        float: Estimated fee rate in satoshis per byte.
        """

        # Get the simplified JSON representation of the PSBT
        simple_psbt = SimplePSBT.from_psbt(psbt)

        # for the input where i can determine the (m,n) use them:
        full_input_mn_tuples = [inp._get_m_of_n() for inp in simple_psbt.inputs]
        input_mn_tuples = [mn for mn in full_input_mn_tuples if mn]

        # Estimate the size of the transaction
        # This part requires the transaction size estimation logic, which might need information about inputs and outputs
        # For simplicity, let's assume you have a function estimate_tx_size(psbt_data) that can estimate the size
        vsize = weight_to_vsize(estimate_tx_weight(input_mn_tuples, len(psbt.extract_tx().output())))

        return FeeInfo(psbt.fee_amount(), vsize, is_estimated=True)

    @classmethod
    def estimate_from_num_inputs(
        cls,
        fee_rate: float,
        input_mn_tuples: List[Tuple[int, int]],
        num_outputs: int,
        include_signatures=True,
    ) -> "FeeInfo":
        "Estimation for segwit txs"
        vsize = weight_to_vsize(
            estimate_tx_weight(
                input_mn_tuples=input_mn_tuples,
                num_outputs=num_outputs,
                include_signatures=include_signatures,
            )
        )
        return FeeInfo(ceil(fee_rate * vsize), vsize, is_estimated=True)


from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional


class PubKeyInfo:
    def __init__(
        self,
        fingerprint: str,
        pubkey: Optional[str] = None,
        derivation_path: Optional[str] = None,
        label: str = "",
    ) -> None:
        self.fingerprint = SimplePubKeyProvider.format_fingerprint(fingerprint)
        self.pubkey = pubkey
        self.derivation_path = derivation_path
        self.label = label


@dataclass
class SimpleInput:
    txin: bdk.TxIn
    witness_script: Optional[str] = None
    # partial_sigs example: {"0232397cde66eb78039694c8c356272d0ea71e621abbaf74f068c16ef5ad5435a6": "304402202976966c2996b3c17005342eebb60133741286460b135020dcb4f2089629d851022044402ada00dc6acb4040f44099f537a138f4558cb10f874656d9d544a19b0f6e"}
    # {pubkey: signature}
    partial_sigs: Dict[str, str] = field(default_factory=dict)
    final_script_sig: Optional[str] = None
    final_script_witness: Optional[str] = None
    pubkeys: List[PubKeyInfo] = field(default_factory=list)
    wallet_id: Optional[str] = None
    m_of_n: Optional[Tuple[int, int]] = None
    non_witness_utxo: Optional[Dict[str, Any]] = None
    witness_utxo: Optional[Dict[str, Any]] = None
    sighash_type: Optional[int] = None
    redeem_script: Optional[str] = None
    ripemd160_preimages: Dict[str, str] = field(default_factory=dict)
    sha256_preimages: Dict[str, str] = field(default_factory=dict)
    hash160_preimages: Dict[str, str] = field(default_factory=dict)
    hash256_preimages: Dict[str, str] = field(default_factory=dict)
    tap_key_sig: Optional[str] = None
    tap_script_sigs: Dict[str, str] = field(default_factory=dict)
    tap_scripts: List[Tuple[str, str, str]] = field(default_factory=list)
    tap_key_origins: List[Tuple[str, List[str]]] = field(default_factory=list)
    tap_internal_key: Optional[str] = None
    tap_merkle_root: Optional[str] = None

    @classmethod
    def from_input(cls, input_data: Dict[str, Any], txin: bdk.TxIn) -> "SimpleInput":
        self = cls(
            txin,
            witness_script=input_data.get("witness_script"),
            partial_sigs={k: v.get("sig") for k, v in input_data.get("partial_sigs", {}).items()},
            final_script_sig=input_data.get("final_script_sig"),
            final_script_witness=input_data.get("final_script_witness"),
            non_witness_utxo=input_data.get("non_witness_utxo"),
            witness_utxo=input_data.get("witness_utxo"),
            sighash_type=input_data.get("sighash_type"),
            redeem_script=input_data.get("redeem_script"),
            ripemd160_preimages=input_data.get("ripemd160_preimages", {}),
            sha256_preimages=input_data.get("sha256_preimages", {}),
            hash160_preimages=input_data.get("hash160_preimages", {}),
            hash256_preimages=input_data.get("hash256_preimages", {}),
            tap_key_sig=input_data.get("tap_key_sig"),
            tap_script_sigs=input_data.get("tap_script_sigs", {}),
            tap_scripts=input_data.get("tap_scripts", []),
            tap_key_origins=input_data.get("tap_key_origins", []),
            tap_internal_key=input_data.get("tap_internal_key"),
            tap_merkle_root=input_data.get("tap_merkle_root"),
        )

        bip32_derivation = input_data.get("bip32_derivation", [])
        for pubkey_info in bip32_derivation:
            pubkey, (fingerprint, derivation_path) = pubkey_info
            self.pubkeys.append(
                PubKeyInfo(pubkey=pubkey, fingerprint=fingerprint, derivation_path=derivation_path)
            )

        self.m_of_n = self._get_m_of_n()
        return self

    def is_fully_signed(self) -> bool:
        # This heuristic assumes the presence of final_script_sig or final_script_witness indicates full signing
        return bool(self.final_script_sig or self.final_script_witness)

    def signature_count(self) -> str:
        # Just return the count of partial signatures, as we can't determine the total required
        return str(len(self.partial_sigs))

    def get_pub_keys_without_signature(self) -> List[PubKeyInfo]:
        # If the input is fully signed, there are no pubkeys without a signature
        if self.is_fully_signed():
            return []
        return [pubkey_info for pubkey_info in self.pubkeys if pubkey_info.pubkey not in self.partial_sigs]

    def get_pub_keys_with_signature(self) -> List[PubKeyInfo]:
        # If the input is fully signed, all pubkeys are considered to have a signature
        if self.is_fully_signed():
            return self.pubkeys
        return [pubkey_info for pubkey_info in self.pubkeys if pubkey_info.pubkey in self.partial_sigs]

    def fingerprint_has_signature(self, fingerprint: str) -> bool:
        if self.is_fully_signed():
            return True
        for pubkey_info in self.pubkeys:
            if pubkey_info.fingerprint == fingerprint and pubkey_info.pubkey in self.partial_sigs:
                return True
        return False

    def get_estimated_m_of_n(self) -> Tuple[int, int]:
        mn = self._get_m_of_n()
        if mn:
            return mn
        return (len(self.pubkeys), len(self.pubkeys))

    def _get_m_of_n(self) -> Optional[Tuple[int, int]]:
        if self.m_of_n:
            return self.m_of_n

        if self.witness_script:
            m, public_keys = parse_witness_script(self.witness_script)
            return (m, len(public_keys))
        if self.redeem_script:
            m, public_keys = parse_redeem_script(self.redeem_script)
            return (m, len(public_keys))
        return None

    def get_prev_txouts(self) -> Dict[str, TxOut]:
        "Returns {str(outpoint): List[TxOut]}"
        if not self.non_witness_utxo:
            return {}
        prev_out = OutPoint.from_bdk(self.txin.previous_output)
        output = self.non_witness_utxo.get("output", [])[prev_out.vout]
        non_witness_utxo_prev_out = TxOut(
            value=output["value"], script_pubkey=bdk.Script(bytes.fromhex(output["script_pubkey"]))
        )
        return {str(prev_out): non_witness_utxo_prev_out}


@dataclass
class SimpleOutput:
    value: int = 0
    script_pubkey: str = ""
    witness_script: Optional[str] = None
    redeem_script: Optional[str] = None
    bip32_derivation: List[PubKeyInfo] = field(default_factory=list)
    tap_internal_key: Optional[str] = None
    tap_tree: Optional[str] = None
    tap_key_origins: List[Tuple[str, List[str]]] = field(default_factory=list)

    @classmethod
    def from_output(cls, output_data: Dict[str, Any], unsigned_tx: Dict[str, Any]) -> "SimpleOutput":
        instance = cls(
            value=unsigned_tx.get("value", 0),
            script_pubkey=unsigned_tx.get("script_pubkey", ""),
            witness_script=output_data.get("witness_script"),
            redeem_script=output_data.get("redeem_script"),
            tap_internal_key=output_data.get("tap_internal_key"),
            tap_tree=output_data.get("tap_tree"),
        )

        bip32_derivation = output_data.get("bip32_derivation", [])
        for pubkey_info in bip32_derivation:
            pubkey, (fingerprint, derivation_path) = pubkey_info
            instance.bip32_derivation.append(
                PubKeyInfo(pubkey=pubkey, fingerprint=fingerprint, derivation_path=derivation_path)
            )

        return instance

    def to_txout(self) -> TxOut:
        return TxOut(self.value, self.script_pubkey)


@dataclass
class SimplePSBT:
    inputs: List[SimpleInput] = field(default_factory=list)
    outputs: List[SimpleOutput] = field(default_factory=list)

    @classmethod
    def from_psbt(cls, psbt: bdk.PartiallySignedTransaction) -> "SimplePSBT":
        instance = cls()
        psbt_json = json.loads(psbt.json_serialize())
        instance.inputs = [
            SimpleInput.from_input(input_data, txin)
            for input_data, txin in zip(psbt_json.get("inputs", []), psbt.extract_tx().input())
        ]

        outputs = psbt_json.get("outputs", [])
        unsigned_tx_outputs = psbt_json.get("unsigned_tx", {}).get("output", [])
        assert len(outputs) == len(unsigned_tx_outputs)
        instance.outputs = [
            SimpleOutput.from_output(output_data, unsigned_tx_output)
            for output_data, unsigned_tx_output in zip(outputs, unsigned_tx_outputs)
        ]
        return instance

    def get_fingerprint_tuples(self) -> Tuple[List[PubKeyInfo], List[PubKeyInfo]]:

        # set of all fingerprints of all inputs
        # sorted_pubkeys = remove_duplicates_keep_order(
        #     sum(
        #         [simple_input.pubkeys for simple_input in self.inputs ],
        #         [],
        #     )
        # )

        pubkeys_not_fully_signed = remove_duplicates_keep_order(
            sum(
                [simple_input.get_pub_keys_without_signature() for simple_input in self.inputs],
                [],
            )
        )
        pubkeys_fully_signed = remove_duplicates_keep_order(
            sum(
                [simple_input.get_pub_keys_with_signature() for simple_input in self.inputs],
                [],
            )
        )

        return (
            # sorted_pubkeys,
            pubkeys_fully_signed,
            pubkeys_not_fully_signed,
        )

    def get_prev_txouts(self) -> Dict[str, TxOut]:
        d = self.inputs[0].get_prev_txouts()
        for inp in self.inputs[1:]:
            d.update(inp.get_prev_txouts())

        return d



import sys

from .dynamic_lib_load import ensure_pyzbar_works

# this setsup the logging
from .logging_setup import setup_logging  # type: ignore

ensure_pyzbar_works()

import argparse
import sys

from PyQt6.QtWidgets import QApplication

from .gui.qt.main import MainWindow
from .gui.qt.util import custom_exception_handler


def parse_args() -> argparse.Namespace:

    parser = argparse.ArgumentParser(description="Bitcoin Safe")
    parser.add_argument("--network", help="Choose the network: bitcoin, regtest, testnet, signet ")
    parser.add_argument(
        "--profile", action="store_true", help="Enable profiling. VIsualize with snakeviz .prof_stats"
    )

    return parser.parse_args()


def main() -> None:
    args = parse_args()

    sys.excepthook = custom_exception_handler
    app = QApplication(sys.argv)
    window = MainWindow(**vars(args))
    window.show()
    app.exec()


if __name__ == "__main__":
    args = parse_args()

    if args.profile:
        import cProfile
        import os
        from pstats import Stats

        with cProfile.Profile() as pr:
            main()

        # run in bash "snakeviz .prof_stats &"  to visualize the stats
        with open("profiling_stats.txt", "w") as stream:
            stats = Stats(pr, stream=stream)
            stats.strip_dirs()
            stats.sort_stats("time")
            stats.dump_stats(".prof_stats")
            os.system("snakeviz .prof_stats & ")
            # os.system("pyprof2calltree -i .prof_stats -k & ")
    else:
        main()



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from datetime import datetime

from bitcoin_safe.util import (
    jsonlines_to_list_of_dict,
    list_of_dict_to_jsonline_list,
    list_of_dict_to_jsonlines,
)

logger = logging.getLogger(__name__)
import copy
import json
from typing import Any, Dict, List, Literal, Union

from bitcoin_qr_tools.data import Data, DataType
from packaging import version

from .pythonbdk_types import *
from .storage import BaseSaveableClass, filtered_for_init

# see https://github.com/bitcoin/bips/blob/master/bip-0329.mediawiki


class Key(enum.Enum):
    type = enum.auto()
    ref = enum.auto()
    label = enum.auto()
    origin = enum.auto()
    spendable = enum.auto()
    category = enum.auto()
    timestamp = enum.auto()


class LabelType:
    tx = "tx"
    addr = "addr"
    pubkey = "pubkey"
    input = "input"
    output = "output"
    xpub = "xpub"


class Label(SaveAllClass):
    VERSION = "0.0.2"
    known_classes = {
        **BaseSaveableClass.known_classes,
    }
    bip329_keys = ["type", "ref", "label"]

    def __init__(
        self,
        type: str,
        ref: str,
        timestamp: float,
        label: Optional[str] = None,
        origin: Optional[str] = None,
        spendable: Optional[bool] = None,
        category: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.type = type
        self.ref = ref
        self.label = label
        self.origin = origin
        self.spendable = spendable
        self.category = category
        self.timestamp = timestamp

    def dump(self, preserve_bip329_keys=True) -> Dict:
        d = super().dump()

        # remove the  key:value with value==none
        for key in list(d.keys()):
            if preserve_bip329_keys and key in self.bip329_keys:
                continue
            if d[key] is None:
                del d[key]

        return d

    @classmethod
    def from_dump_migration(cls, dct: Dict[str, Any]) -> Dict[str, Any]:
        if version.parse(str(dct["VERSION"])) <= version.parse("0.0.0"):
            pass

        if version.parse(str(dct["VERSION"])) <= version.parse("0.0.1"):
            if "flat_data" in dct:
                #
                dct["timestamp"] = dct["timestamp"] if dct["timestamp"] else datetime.now().timestamp()

        # now the version is newest, so it can be deleted from the dict
        if "VERSION" in dct:
            del dct["VERSION"]
        return dct

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Label):
            return False
        return self.dump() == other.dump()


class Labels(BaseSaveableClass):
    VERSION = "0.1.0"
    known_classes = {**BaseSaveableClass.known_classes, "Label": Label}

    def __init__(
        self,
        data: Dict[str, Label] | None = None,
        categories: Optional[List[str]] = None,
        default_category: str = "default",
    ) -> None:
        super().__init__()

        # "tb1q6xhxcrzmjwf6ce5jlj08gyrmu4eq3zwpv0ss3f":{ "type": "addr", "ref": "tb1q6xhxcrzmjwf6ce5jlj08gyrmu4eq3zwpv0ss3f", "label": "Address" }
        self.data: Dict[str, Label] = data if data else {}
        self.categories: List[str] = categories if categories else []

        self.separator = " #"
        self.default_category = default_category

    def add_category(self, value: str) -> None:
        if value not in self.categories:
            self.categories.append(value)

    def del_item(self, ref: str) -> None:
        if ref in self.data:
            del self.data[ref]

    def get_label(self, ref: str, default_value: str | None = None) -> Optional[str]:
        item = self.data.get(ref)
        if not item:
            return default_value
        return item.label

    def get_category_raw(self, ref: str) -> str | None:
        item = self.data.get(ref)
        if not item or item.category is None:
            return None

        return item.category

    def get_category(self, ref: str, default_value=None) -> str:
        item = self.data.get(ref)
        if not item or item.category is None:
            return default_value if default_value else self.get_default_category()

        return item.category

    def get_timestamp(self, ref: str, default_value=None) -> Optional[float]:
        item = self.data.get(ref)
        if not item or not item.timestamp:
            return default_value

        return item.timestamp

    def set_label(
        self, type: str, ref: str, label_value, timestamp: Union[Literal["now"], float] = "now"
    ) -> None:
        label = self.data.get(ref)
        timestamp = datetime.now().timestamp() if timestamp == "now" else timestamp
        if not label:
            self.data[ref] = label = Label(type, ref, timestamp)

        label.label = label_value
        label.timestamp = timestamp

        if all(value is None for value in [label.category, label.spendable, label.label, label.origin]):
            del self.data[ref]

    def set_category(
        self, type: str, ref: str, category, timestamp: Union[Literal["now"], float] = "now"
    ) -> None:
        label = self.data.get(ref)
        timestamp = datetime.now().timestamp() if timestamp == "now" else timestamp
        if not label:
            self.data[ref] = label = Label(type, ref, timestamp)

        label.category = category
        label.timestamp = timestamp

        if all(value is None for value in [label.category, label.spendable, label.label, label.origin]):
            del self.data[ref]

        if category and category not in self.categories:
            self.categories.append(category)

    def set_tx_label(self, label_value, value, timestamp: Union[Literal["now"], float] = "now") -> None:
        return self.set_label(LabelType.tx, label_value, value, timestamp=timestamp)

    def set_addr_label(self, ref: str, label_value, timestamp: Union[Literal["now"], float] = "now") -> None:
        return self.set_label(LabelType.addr, ref, label_value, timestamp=timestamp)

    def set_pubkey_label(
        self, ref: str, label_value, timestamp: Union[Literal["now"], float] = "now"
    ) -> None:
        return self.set_label(LabelType.pubkey, ref, label_value, timestamp=timestamp)

    def set_input_label(self, ref: str, label_value, timestamp: Union[Literal["now"], float] = "now") -> None:
        return self.set_label(LabelType.input, ref, label_value, timestamp=timestamp)

    def set_output_label(
        self, ref: str, label_value, timestamp: Union[Literal["now"], float] = "now"
    ) -> None:
        return self.set_label(LabelType.output, ref, label_value, timestamp=timestamp)

    def set_xpub_label(self, ref: str, label_value, timestamp: Union[Literal["now"], float] = "now") -> None:
        return self.set_label(LabelType.xpub, ref, label_value, timestamp=timestamp)

    def set_addr_category(self, ref: str, category, timestamp: Union[Literal["now"], float] = "now") -> None:
        return self.set_category(LabelType.addr, ref, category, timestamp=timestamp)

    def set_tx_category(self, ref: str, category, timestamp: Union[Literal["now"], float] = "now") -> None:
        return self.set_category(LabelType.tx, ref, category, timestamp=timestamp)

    def get_default_category(self) -> str:
        return self.categories[0] if self.categories else self.default_category

    def dump(self) -> Dict:
        d = super().dump()

        d["data"] = self.data
        d["default_category"] = self.default_category

        keys = ["categories"]
        for k in keys:
            d[k] = copy.deepcopy(self.__dict__[k])
        return d

    @classmethod
    def from_dump(cls, dct: Dict, class_kwargs=None) -> "Labels":
        super()._from_dump(dct, class_kwargs=class_kwargs)

        return cls(**filtered_for_init(dct, cls))

    @classmethod
    def from_dump_migration(cls, dct: Dict[str, Any]) -> Dict[str, Any]:
        "this class should be overwritten in child classes"
        if version.parse(str(dct["VERSION"])) <= version.parse("0.0.0"):
            if "data" in dct:
                #
                dct["flat_data"] = list(dct["data"].values())
        if version.parse(str(dct["VERSION"])) <= version.parse("0.0.1"):
            if "flat_data" in dct:
                #
                del dct["flat_data"]
        if version.parse(str(dct["VERSION"])) <= version.parse("0.0.4"):
            if "data" in dct:
                #
                dct["data"] = {
                    k: Label(**v, timestamp=datetime.now().timestamp()) for k, v in dct["data"].items()
                }

        # now the VERSION is newest, so it can be deleted from the dict
        if "VERSION" in dct:
            del dct["VERSION"]
        return dct

    def _convert_item_to_bip329(self, label: Label) -> Dict:

        new_label: Label = label.clone()

        if new_label.category:
            new_label.label = (
                f'{new_label.label if new_label.label else ""}{self.separator}{new_label.category}'
            )
            new_label.category = None

        dct = new_label.dump()
        del dct["__class__"]
        del dct["VERSION"]
        del dct["timestamp"]
        return dct

    def _bip329_dict_to_label(
        self, d: Dict[str, Any], timestamp: Union[Literal["now"], float] = "now"
    ) -> Label:
        d["timestamp"] = datetime.now().timestamp() if timestamp == "now" else timestamp
        label = Label(**d)

        if label.label and (not label.category) and self.separator in label.label:
            label.label, *categories = label.label.split(self.separator)
            if categories:
                if len(categories) > 1:
                    logger.warning(f"categories = {categories}. Dropping all but the first non-empty.")

                for category in categories:
                    # clean category
                    category = category.replace(self.separator.strip(), "").strip()
                    if not category:
                        continue
                    label.category = category
                    break
        return label

    def export_bip329_jsonlines(self) -> str:
        list_of_dict = [self._convert_item_to_bip329(item) for item in self.data.values()]
        return list_of_dict_to_jsonlines(list_of_dict)

    def import_bip329_jsonlines(
        self, jsonlines: str, fill_categories=True, timestamp: Union[Literal["now"], float] = "now"
    ) -> Dict[str, Label]:
        list_of_dict = jsonlines_to_list_of_dict(jsonlines)
        labels = [self._bip329_dict_to_label(d, timestamp=timestamp) for d in list_of_dict]
        return self.import_labels(labels=labels, fill_categories=fill_categories)

    def import_electrum_wallet_json(
        self,
        file_content: str,
        network: bdk.Network,
        fill_categories=True,
        timestamp: Union[Literal["now"], float] = "now",
    ) -> Dict[str, Label]:
        electrum_dict = json.loads(file_content)
        list_of_dict = []
        for key, label in electrum_dict.items():
            data = Data.from_str(key, network)
            data_type = "addr"
            if data.data_type == DataType.Txid:
                data_type = "tx"
            list_of_dict.append({"type": data_type, "ref": key, "label": label})

        labels = [self._bip329_dict_to_label(d, timestamp=timestamp) for d in list_of_dict]
        return self.import_labels(labels=labels, fill_categories=fill_categories)

    def _should_overwrite(self, new_label: Label, old_label: Optional[Label]) -> bool:
        if not old_label:
            return True
        if not new_label.timestamp:
            return False
        if not old_label.timestamp:
            return True
        return new_label.timestamp > old_label.timestamp

    def import_labels(
        self, labels: List[Label], fill_categories=True, force_overwrite=False
    ) -> Dict[str, Label]:
        changed_data: Dict[str, Label] = {}

        for label in labels:
            if self.data.get(label.ref) != label and (
                force_overwrite or self._should_overwrite(new_label=label, old_label=self.data.get(label.ref))
            ):
                self.data[label.ref] = label
                changed_data[label.ref] = label

        if fill_categories:
            for item in self.data.values():
                if item.category is None:
                    continue
                if item.category not in self.categories:
                    self.categories.append(item.category)
        return changed_data

    def dumps_data_jsonline_list(self, refs: list[str] | None = None) -> List[str]:
        return list_of_dict_to_jsonline_list(
            [label.dump() for ref, label in self.data.items() if (refs is None) or (ref in refs)]
        )

    def dumps_data_jsonlines(self, refs: list[str] | None = None) -> str:
        return list_of_dict_to_jsonlines(
            [label.dump() for ref, label in self.data.items() if (refs is None) or (ref in refs)]
        )

    def import_dumps_data(
        self, dumps_data: str, fill_categories=True, force_overwrite=False
    ) -> Dict[str, Label]:
        labels = [Label.from_dump(label_dict) for label_dict in jsonlines_to_list_of_dict(dumps_data)]
        return self.import_labels(
            labels=labels, fill_categories=fill_categories, force_overwrite=force_overwrite
        )

    def rename_category(self, old_category: str, new_category: str) -> List[str]:
        affected_keys: List[str] = []
        for key, item in list(self.data.items()):
            if item.category and item.category == old_category:
                item.category = new_category
                item.timestamp = datetime.now().timestamp()
                affected_keys.append(key)

        if old_category in self.categories:
            idx = self.categories.index(old_category)
            self.categories.pop(idx)
            if new_category not in self.categories:
                self.categories.insert(idx, new_category)
        return affected_keys

    def delete_category(self, category: str) -> List[str]:
        affected_keys = []
        for key, item in list(self.data.items()):
            if item.category and item.category == category:
                affected_keys.append(key)
                item.category = self.get_default_category()
                item.timestamp = datetime.now().timestamp()

        if category in self.categories:
            idx = self.categories.index(category)
            self.categories.pop(idx)

        return affected_keys
